#include <stdio.h>
#include <proto.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include "gtapp.h"
#include "gtappdisp.h"
#include "gtwidget.h"
#include "gtconfig.h"
#include "gtcommon.h"
#include "gterror.h"

/************************************************************************\
*
* GtApp class.
*
* $Ident: This file is part of the 'Gt-library' 1.0.0 $
* $Copyright: (C) 1997 by Gran Gustafsson $
*
\************************************************************************/

#define FMAXL		2048

/* Type of elements on resource converter list */
typedef struct GtAppList
{
	GtDLink lnk;
	long info;
	GtCallB Fun;
	GtUDPtr ud;
} GtAppList;

/* Type of elements on timeout list */
typedef struct GtTOList
{
	GtDLink lnk;
	struct timeval tv;
	GtCallB Fun;
	GtUDPtr ud;
} GtTOList;

/* Type of resource type string => resource type number list */
typedef struct GtRTList
{
	GtDLink lnk;
	char *name;
	int nr;
} GtRTList;

/* Type of elements on fd watch list */
typedef struct GtFDWList
{
	GtDLink lnk;
	int fd;
	uword mode;
	GtCallB Fun;
	GtUDPtr ud;
} GtFDWList;

/* Type of elements in display list */
typedef struct GtDspList
{
	GtSLink lnk;
	GtAppDisp *dsp;
	sbool pend;
} GtDspList;

/* Flags for app */
#define GTAF_BREAK		0x01

/* Temp storage for resource file */
static char fbuf[FMAXL];

/* First object of type GtApp */
static sbool fstinit=TRUE;

/* Resources to look for on cmdline */
static XrmOptionDescRec apptbl[]=
{
	{"-background","*background",XrmoptionSepArg,NULL},
	{"-bg","*background",XrmoptionSepArg,NULL},
	{"-foreground","*foregorund",XrmoptionSepArg,NULL},
	{"-fg","*foreground",XrmoptionSepArg,NULL},
	{"-light","*light",XrmoptionSepArg,NULL},
	{"-dark","*dark",XrmoptionSepArg,NULL},
	{"-fill","*fill",XrmoptionSepArg,NULL},
	{"-font","*fontlist",XrmoptionSepArg,NULL},
	{"-fn","*fontlist",XrmoptionSepArg,NULL},
	{"-geometry",".GtAppShell.geometry",XrmoptionSepArg,NULL},
	{"-title",".GtAppShell.title",XrmoptionSepArg,NULL},
	{"-display",".dispstring",XrmoptionSepArg,NULL},
	{"-iconic","GtAppShell.iconic",XrmoptionNoArg,"true"},
	{"-xrm",NULL,XrmoptionResArg,NULL}
};
#define XRMTBLNR  14

/************************************************************************\
*
* Convert from string to number.
*
\************************************************************************/
static int GtConvNum(GtBase *obj,GtCDPtr cd,GtUDPtr ud)
{
	GtResourceCD *rcd;
	
	obj=obj;
	ud=ud;
	rcd=(GtResourceCD *)cd;
	switch(rcd->type)
	{
		case GTRT_BYTE:
			*((byte *)rcd->data)=(byte)strtol(rcd->res.addr,NULL,0);
			break;
		case GTRT_WORD:
			*((word *)rcd->data)=(word)strtol(rcd->res.addr,NULL,0);
			break;
		case GTRT_LONG:
			*((long *)rcd->data)=(long)strtol(rcd->res.addr,NULL,0);
			break;
	}
	return(GTCB_BREAK);
}

/************************************************************************\
*
* Convert from string to named constant.
*
\************************************************************************/
static int GtConvStrs(GtBase *obj,GtCDPtr cd,GtUDPtr ud)
{
	ubyte v=0;
	char *str;
	GtResourceCD *rcd;
	
	obj=obj;
	ud=ud;
	rcd=(GtResourceCD *)cd;
	str=(char *)rcd->res.addr;
	switch(rcd->type)
	{
		case GTRT_ALIGN:
			if(strcmp(str,"left")==0)
				v=GTA_LEFT;
			else if(strcmp(str,"right")==0)
				v=GTA_RIGHT;
			else if(strcmp(str,"center")==0)
				v=GTA_CENTER;
			else
				return(GTCB_CONT);
			break;
		case GTRT_PLACE:
			if((strcmp(str,"topleft")==0) || (strcmp(str,"top")==0) ||
					(strcmp(str,"left")==0))
				v=GTP_TOPLEFT;
			else if((strcmp(str,"botright")==0) || (strcmp(str,"bottom")==0) ||
					(strcmp(str,"right")==0))
				v=GTP_BOTRIGHT;
			else if(strcmp(str,"center")==0)
				v=GTP_CENTER;
			else if(strcmp(str,"expand")==0)
				v=GTP_EXPAND;
			else
				return(GTCB_CONT);
			break;
		case GTRT_ORIENT:
			if((strcmp(str,"horiz")==0) || (strcmp(str,"horizontal")==0))
				v=GTO_HORIZ;
			else if((strcmp(str,"vert")==0) || (strcmp(str,"vertical")==0))
				v=GTO_VERT;
			else
				return(GTCB_CONT);
			break;
		case GTRT_DIR:
			if(strcmp(str,"left")==0)
				v=GTD_LEFT;
			else if(strcmp(str,"right")==0)
				v=GTD_RIGHT;
			else if(strcmp(str,"up")==0)
				v=GTD_UP;
			else if(strcmp(str,"down")==0)
				v=GTD_DOWN;
			else
				return(GTCB_CONT);
			break;
		case GTRT_BOOL:
			if(strcmp(str,"true")==0)
				v=TRUE;
			else if(strcmp(str,"false")==0)
				v=FALSE;
			else
				return(GTCB_CONT);
			break;
		case GTRT_FRAME:
			if(strcmp(str,"none")==0)
				v=GTF_NONE;
			else if(strcmp(str,"simple")==0)
				v=GTF_SIMPLE;
			else if(strcmp(str,"dbl_in")==0)
				v=GTF_DBL_IN;
			else if(strcmp(str,"dbl_out")==0)
				v=GTF_DBL_OUT;
			else if(strcmp(str,"in")==0)
				v=GTF_IN;
			else if(strcmp(str,"out")==0)
				v=GTF_OUT;
			else
				return(GTCB_CONT);
			break;
		case GTRT_APLACE:
			if(strcmp(str,"apart")==0)
				v=GTAP_APART;
			else if(strcmp(str,"together")==0)
				v=GTAP_TOGETHER;
			else
				return(GTCB_CONT);
			break;
	}
	*((ubyte *)rcd->data)=v;
	return(GTCB_BREAK);
}

/************************************************************************\
*
* Create and init app.
*
\************************************************************************/
GtApp *GtMkApp(GtTag *tgs)
{
	GtApp *obj;
	
	if(!(obj=new GtApp()))
	{
		GtError(NULL,GTEC_NOOBJECT,"No GtApp object!\n");
		return(NULL);
	}
	if(!obj->Init(tgs))
	{
		GtDestroy(obj);
		obj=NULL;
	}
	return(obj);
}
	
/************************************************************************\
*
* Create and init app.
*
\************************************************************************/
GtApp *GtMkApp(ulong tag,...)
{
	va_list va;
	GtApp *obj;
	int p;
	
	va_start(va,tag);
	p=GtBeginVaTags();
	obj=GtMkApp(GtCopyVaTags(tag,&va));
	GtEndVaTags(p);
	va_end(va);
	return(obj);
}
	
/************************************************************************\
*
* Setup empty app.
*
\************************************************************************/
GtApp::GtApp(void)
{
	GtDInit(&tohd);
	GtDInit(&rchd);
	GtDInit(&ftohd);
	GtDInit(&rthd);
	GtDInit(&fdhd);
	GtSInit(&disps);
	robj=NULL;
	aname=NULL;
	cname=NULL;
	nprefix=NULL;
	cprefix=NULL;
	nplen=0;
	cplen=0;
	npmlen=0;
	cpmlen=0;
	argc=0;
	argv=NULL;
	nxtrtype=GTRT_LAST_TYPE;
	FD_ZERO(&mrd);
	FD_ZERO(&mwr);
	FD_ZERO(&mex);
	flags=0;
}

/************************************************************************\
*
* Destroy app.
*
\************************************************************************/
void GtApp::Destroy(void)
{
	GtAppList *al;
	GtTOList *to;
	GtDspList *dl;
	
	while((dl=(GtDspList *)GtSFirst(&disps)))
	{
		/* Destroy display and it will remove itself from the list */
		GtDestroy(dl->dsp);
	}
	if(appdb)
		XrmDestroyDatabase(appdb);
	if(nprefix)
		free(nprefix);
	if(cprefix)
		free(cprefix);
	while((to=(GtTOList *)GtDRemHead(&tohd)))
		free(to);	
	while((al=(GtAppList *)GtDRemHead(&rchd)))
		free(al);
	GtBase::Destroy();
}

/************************************************************************\
*
* Setup app.
*
\************************************************************************/
sbool GtApp::Init(GtTag *tgs)
{
	GtTag *t;
	char *rpath,*p;
	char *str;

	while((t=GtNextTag(&tgs)))
	{
		switch(t->tag)
		{
			case GT_AppName:
				aname=(char *)t->d.ptr;
				break;
			case GT_AppClass:
				cname=(char *)t->d.ptr;
				break;
		}
	}
	if(!aname)
		aname="gtapp";
	if(!cname)
		cname="GtApp";
	Prefix(NULL,0,FALSE);
	Prefix(NULL,0,TRUE);
	if(fstinit)
	{
		fstinit=FALSE;
		XrmInitialize();
		appdb=NULL;
		if(!(rpath=strdup(GTDEF_RESPATH)))
		{
			GtError(this,GTEC_NOMEMORY,"No memory for resource paths!\n");
			return(FALSE);
		}
		for(p=strtok(rpath,":"); p; p=strtok(NULL,":"))
		{
			sprintf(fbuf,p,GTDEF_DEFAULTS);
			XrmCombineFileDatabase(fbuf,&appdb,True);
			sprintf(fbuf,p,cname);
			XrmCombineFileDatabase(fbuf,&appdb,True);
			sprintf(fbuf,p,aname);
			XrmCombineFileDatabase(fbuf,&appdb,True);
		}
		free(rpath);
		if((str=getenv("GTRESOURCEPATH")))
		{
			if(!(rpath=strdup(str)))
			{
				GtError(this,GTEC_NOMEMORY,"No memory for resource paths!\n");
				return(FALSE);
			}
			for(p=strtok(rpath,":"); p; p=strtok(NULL,":"))
			{
				sprintf(fbuf,p,GTDEF_DEFAULTS);
				XrmCombineFileDatabase(fbuf,&appdb,True);
				sprintf(fbuf,p,cname);
				XrmCombineFileDatabase(fbuf,&appdb,True);
				sprintf(fbuf,p,aname);
				XrmCombineFileDatabase(fbuf,&appdb,True);
			}
			free(rpath);
		}
		AddRConverter(GTRT_BYTE,GtConvNum,NULL);
		AddRConverter(GTRT_WORD,GtConvNum,NULL);
		AddRConverter(GTRT_LONG,GtConvNum,NULL);
		AddRConverter(GTRT_ALIGN,GtConvStrs,NULL);
		AddRConverter(GTRT_PLACE,GtConvStrs,NULL);
		AddRConverter(GTRT_ORIENT,GtConvStrs,NULL);
		AddRConverter(GTRT_FRAME,GtConvStrs,NULL);
		AddRConverter(GTRT_DIR,GtConvStrs,NULL);
		AddRConverter(GTRT_BOOL,GtConvStrs,NULL);
		AddRConverter(GTRT_APLACE,GtConvStrs,NULL);
	}
	return(TRUE);
}

/************************************************************************\
*
* Set app attributes.
*
\************************************************************************/
sbool GtApp::Set(GtTag *tgs)
{
	tgs=tgs;
	return(TRUE);
}

/************************************************************************\
*
* Get app attributes.
*
\************************************************************************/
void GtApp::Get(GtTag *tgs)
{
	GtTag *t;
	long *ptr;
	
	while((t=GtNextTag(&tgs)))
	{
		if(!(ptr=(long *)t->d.ptr))
			continue;
		switch(t->tag)
		{
			case GT_AppName:
				*ptr=(long)aname;
				break;
			case GT_AppClass:
				*ptr=(long)cname;
				break;
		}
	}
}

/************************************************************************\
*
* Check if class is (sub)class of named class.
*
\************************************************************************/
sbool GtApp::IsSubClassOf(char *cn)
{
	if(strcmp(cn,"GtApp")==0)
		return(TRUE);
	else
		return(GtBase::IsSubClassOf(cn));
}

/************************************************************************\
*
* Setup resources from command line.
*
\************************************************************************/
void GtApp::SetupRes(int *ac,char **av)
{
	XrmParseCommand(&appdb,apptbl,XRMTBLNR,aname,ac,av);
	argc=*ac;
	argv=av;
}

/************************************************************************\
*
* Load resource file.
*
\************************************************************************/
sbool GtApp::LoadRes(char *name,sbool override)
{
	int ret;
	
	if(override)
		ret=XrmCombineFileDatabase(name,&appdb,True);
	else
		ret=XrmCombineFileDatabase(name,&appdb,False);
	if(ret)
		return(TRUE);
	return(FALSE);
}

/************************************************************************\
*
* Get and convert resource.
*
\************************************************************************/
sbool GtApp::GetRes(GtWidget *obj,char *name,int type,void *data)
{
	GtAppList *al;
	char *dmy;

	if(appdb || (obj && obj->GetDisp()->GetDB()))
	{
		if(obj!=robj)
		{
			Prefix(obj,0,TRUE);
			Prefix(obj,0,FALSE);
		}
		robj=obj;
		if(cprefix && nprefix)
		{
			strncpy(nprefix+nplen,name,GTDEF_MAXRESLEN);
			strncpy(cprefix+cplen,name,GTDEF_MAXRESLEN);
			nprefix[GTDEF_MAXRESLEN]='\0';
			cprefix[GTDEF_MAXRESLEN]='\0';
			if(XrmGetResource(appdb,nprefix,cprefix,&dmy,&rval.res) || (obj && 
				XrmGetResource(obj->GetDisp()->GetDB(),nprefix,cprefix,&dmy,&rval.res)))
			{
				if(type!=GTRT_STRING)
				{
					rval.type=type;
					rval.data=data;
					for(al=(GtAppList *)GtDFirst(&rchd); al;
							al=(GtAppList *)GtDNext(&rchd,&al->lnk))
					{
						if(al->info==type)
						{
							if(al->Fun(robj,&rval,al->ud)==GTCB_BREAK)
								return(TRUE);
						}
					}
					return(FALSE);
				}
				*((char **)data)=(char *)rval.res.addr;
				return(TRUE);
			}
		}
	}
	return(FALSE);
}

/************************************************************************\
*
* Add resource converter.
*
\************************************************************************/
void GtApp::AddRConverter(int type,GtCallB cb,GtUDPtr ud)
{
	GtAppList *al;
	
	for(al=(GtAppList *)GtDFirst(&rchd); al;
			al=(GtAppList *)GtDNext(&rchd,&al->lnk))
	{
		if((al->info==type) && (al->Fun==cb))
			return;
	}
	if(!(al=(GtAppList *)malloc(sizeof(GtAppList))))
	{
		GtError(this,GTEC_NOMEMORY,"No memory for resource converter!\n");
		return;
	}
	al->info=type;
	al->Fun=cb;
	al->ud=ud;
	GtDAddTail(&rchd,&al->lnk);
}

/************************************************************************\
*
* Remove resource converter.
*
\************************************************************************/
void GtApp::RemRConverter(int type,GtCallB cb)
{
	GtAppList *al;
	
	for(al=(GtAppList *)GtDFirst(&rchd); al;
			al=(GtAppList *)GtDNext(&rchd,&al->lnk))
	{
		if((al->info==type) && (al->Fun==cb))
		{
			GtDRemove(&al->lnk);
			free(al);
			return;
		}
	}
}

/************************************************************************\
*
* Register or check for named resource type (for use in converter).
*
\************************************************************************/
int GtApp::AddRType(char *tname,sbool create)
{
	GtRTList *rl;
	
	for(rl=(GtRTList *)GtDFirst(&rthd); rl;
			rl=(GtRTList *)GtDNext(&rthd,&rl->lnk))
	{
		if(strcmp(rl->name,tname)==0)
			return(rl->nr);
	}
	if(!create)
		return(0);
	if(!(rl=(GtRTList *)malloc(sizeof(GtRTList))))
	{
		GtError(this,GTEC_NOMEMORY,"No memory for resource type!\n");
		return(0);
	}
	nxtrtype++;
	rl->name=tname;
	rl->nr=nxtrtype;
	GtDAddTail(&rthd,&rl->lnk);
	return(rl->nr);
}

/************************************************************************\
*
* Add display to app.
*
\************************************************************************/
void GtApp::AddDisp(GtAppDisp *d)
{
	GtDspList *dl;
	
	if(d)
	{
		if(!(dl=(GtDspList *)malloc(sizeof(GtDspList))))
		{
			GtError(this,GTEC_NOMEMORY,"No memory for display list!\n");
			return;
		}
		dl->dsp=d;
		dl->pend=FALSE;
		FD_SET(d->GetFD(),&mrd);
		GtSAddTail(&disps,&dl->lnk);
	}
}

/************************************************************************\
*
* Remove display from app.
*
\************************************************************************/
void GtApp::RemDisp(GtAppDisp *d)
{
	GtDspList *dl;
	
	if(!d)
		return;
	for(dl=(GtDspList *)GtSFirst(&disps); dl;
			dl=(GtDspList *)GtSNext(&dl->lnk))
	{
		if(dl->dsp==d)
		{
			GtSRemove(&disps,&dl->lnk);
			FD_CLR(d->GetFD(),&mrd);
			free(dl);
			return;
		}
	}
}

/************************************************************************\
*
* Add a timeout callback.
*
\************************************************************************/
void GtApp::AddTimeOut(long msec,GtCallB cb,GtUDPtr ud)
{
	GtTOList *to,*tto;
	long tmp;
	
	if(msec<=0)
		return;
	if((to=(GtTOList *)GtDRemHead(&ftohd)) || 
		 (to=(GtTOList *)malloc(sizeof(GtTOList))))
	{
		gettimeofday(&to->tv,NULL);
		tmp=to->tv.tv_usec;
		to->tv.tv_usec+=(msec%1000)*1000;
		to->tv.tv_sec+=(msec/1000);
		if(to->tv.tv_usec<tmp)
			to->tv.tv_sec++;
		to->Fun=cb;
		to->ud=ud;
		for(tto=(GtTOList *)GtDFirst(&tohd); tto; 
			tto=(GtTOList *)GtDNext(&tohd,&tto->lnk))
		{
			if(timercmp(&to->tv,&tto->tv,<))
			{
				GtDAddBefore(&tto->lnk,&to->lnk);
				break;
			}
		}
		if(!tto)
			GtDAddTail(&tohd,&to->lnk);
	}
	else
		GtError(this,GTEC_NOMEMORY,"No memory for timeout!\n");
}
					
/************************************************************************\
*
* Remove a timeout callback.
*
\************************************************************************/
void GtApp::RemTimeOut(GtCallB cb)
{
	GtTOList *to;
	
	for(to=(GtTOList *)GtDFirst(&tohd); to;
			to=(GtTOList *)GtDNext(&tohd,&to->lnk))
	{
		if(to->Fun==cb)
		{
			GtDRemove(&to->lnk);
			GtDAddTail(&ftohd,&to->lnk);
			break;
		}
	}
}

/************************************************************************\
*
* Add file descriptor watch.
*
\************************************************************************/
void GtApp::AddFDWatch(int fd,uword mode,GtCallB cb,GtUDPtr ud)
{
	GtFDWList *fl;
	
	if(!(fl=(GtFDWList *)malloc(sizeof(GtFDWList))))
	{
		GtError(this,GTEC_NOMEMORY,"No memory for file descriptor watch: %d!\n",fd);
		return;
	}
	fl->fd=fd;
	fl->mode=mode;
	fl->Fun=cb;
	fl->ud=ud;
	GtDAddTail(&fdhd,&fl->lnk);
	if(mode & GTAWM_READ)
		FD_SET(fl->fd,&mrd);
	if(mode & GTAWM_WRITE)
		FD_SET(fl->fd,&mwr);
	if(mode & GTAWM_EXCEPT)
		FD_SET(fl->fd,&mex);
}

/************************************************************************\
*
* Remove file descriptor watch.
*
\************************************************************************/
void GtApp::RemFDWatch(int fd,GtCallB cb)
{
	GtFDWList *fl;
	
	for(fl=(GtFDWList *)GtDFirst(&fdhd); fl;
			fl=(GtFDWList *)GtDNext(&fdhd,&fl->lnk))
	{
		if((fl->fd==fd) && (fl->Fun==cb))
		{
			GtDRemove(&fl->lnk);
			free(fl);
			break;
		}
	}
}

/************************************************************************\
*
* Set flag to break infinite event loop.
*
\************************************************************************/
void GtApp::BreakLoop(void)
{
	flags|=GTAF_BREAK;
}

/************************************************************************\
*
* Handle events from X and other sources.
*
\************************************************************************/
sbool GtApp::EventLoop(int mode)
{
	GtDspList *dl;
	fd_set rd,wr,ex;
	struct timeval tv,ct,*t=NULL;
	GtTOList *to;
	GtFDWList *fl,*tmp;
	GtWatchCD cd;
	sbool pnd,ret=TRUE;
	
	while(1)
	{
		// Setup fd sets
		rd=mrd;
		wr=mwr;
		ex=mex;
		// Check for pending X-events in any display
		pnd=FALSE;
		for(dl=(GtDspList *)GtSFirst(&disps); dl;
				dl=(GtDspList *)GtSNext(&dl->lnk))
		{
			if(dl->dsp->Pending())
			{
				// At least one display has pending events
				pnd=TRUE;
				dl->pend=TRUE;
			}
			else
				dl->pend=FALSE;
		}
		// If no pending events and not polled mode calculate sleep time
		if(!pnd && (mode!=GTAM_POLL))
		{
			// If any timeout calculate sleep time from it or else sleep unlimited
			if((to=(GtTOList *)GtDFirst(&tohd)))
			{
				gettimeofday(&ct,NULL);
				MaxSleepTime(&tv,&to->tv,&ct);
				t=&tv;
			}
			else
				t=NULL;
		}
		else
		{
			// Polled mode or pending events so no sleep
			tv.tv_sec=0;
			tv.tv_usec=0;
			t=&tv;
		}
		// Check file descriptors
		select(FD_SETSIZE,&rd,&wr,&ex,t);
		// Check if any timeout
		CheckTimeOut();
		// Check if any file descriptor watches are to be called
		for(fl=(GtFDWList *)GtDFirst(&fdhd); fl; fl=tmp)
		{
			tmp=(GtFDWList *)GtDNext(&fdhd,&fl->lnk);
			cd.fd=fl->fd;
			cd.mode=0;
			if(FD_ISSET(fl->fd,&rd))
				cd.mode|=GTAWM_READ;
			if(FD_ISSET(fl->fd,&wr))
				cd.mode|=GTAWM_WRITE;
			if(FD_ISSET(fl->fd,&ex))
				cd.mode|=GTAWM_EXCEPT;
			if(cd.mode)
			{
				fl->Fun(this,&cd,fl->ud);
			}
		}
		// Loop through all displays and handle events if any pending
		for(dl=(GtDspList *)GtSFirst(&disps); dl;
				dl=(GtDspList *)GtSNext(&dl->lnk))
		{
			if(dl->pend)
				dl->dsp->HandleEvent(NULL);
		}
		// Check if the event loop is to be terminated
		if((mode!=GTAM_LOOP) || (flags & GTAF_BREAK))
		{
			if(flags & GTAF_BREAK)
				ret=FALSE;
			flags&=~GTAF_BREAK;
			break;
		}
	}
	return(ret);
}

/************************************************************************\
*
* Describe this library.
*
\************************************************************************/
char *GtDesc(void)
{
	return("G-Toolkit (Gt) (C)1996 by Gran Gustafsson");
}

/************************************************************************\
*
* Create resource prefix.
*
\************************************************************************/
char *GtApp::Prefix(GtWidget *obj,int sz,sbool cls)
{
	int len;
	char *str;
	
	if(obj)
	{
		// Find length of name or class strings
		if(cls)
			len=strlen(obj->GetClass());
		else
			len=strlen(obj->GetName());
		// Recursive call to find total length and alloc memory
		if((str=Prefix((GtWidget *)obj->GetParent(),sz+(len ? len+1 : 0),cls)))
		{
			if(len)
			{
				// Add name or class to prefix
				if(cls)
					strcat(str,obj->GetClass());
				else
					strcat(str,obj->GetName());
				// Append a '.' to prefix
				strcat(str,".");
			}
		}
	}
	else
	{
		if(cls)
		{
			// Find length of application class
			sz+=strlen(cname);
			// Check if new length is longer than current max length
			if(sz+GTDEF_MAXRESLEN+1>cpmlen)
			{
				cpmlen=sz+GTDEF_MAXRESLEN+1;
				// Extend memory for class prefix
				if(!(cprefix=(char *)realloc(cprefix,cpmlen)))
				{
					GtError(this,GTEC_NOMEMORY,"No memory for resource prefix!\n");
					return(NULL);
				}
			}
			// Add class name prefix
			strcpy(cprefix,cname);
			cplen=sz+1;
			str=cprefix;
		}
		else
		{
			// Find length application name
			sz+=strlen(aname);
			// Check if new length is longer than current max length
			if(sz+GTDEF_MAXRESLEN+1>npmlen)
			{
				npmlen=sz+GTDEF_MAXRESLEN+1;
				// Extend memory for name prefix
				if(!(nprefix=(char *)realloc(nprefix,npmlen)))
				{
					GtError(this,GTEC_NOMEMORY,"No memory for resource prefix!\n");
					return(NULL);
				}
			}
			// Add class name prefix
			strcpy(nprefix,aname);
			nplen=sz+1;
			str=nprefix;
		}
		strcat(str,".");
	}
	return(str);
}

/************************************************************************\
*
* Calculate maximum time to sleep.
*
\************************************************************************/
void GtApp::MaxSleepTime(struct timeval *res,
													struct timeval *x,struct timeval *y)
{
	int nsec;
	
	if(x->tv_usec<y->tv_usec)
	{
		nsec=(y->tv_usec-x->tv_usec)/1000000+1;
		y->tv_usec-=1000000*nsec;
		y->tv_sec+=nsec;
	}
	if(x->tv_usec-y->tv_usec>1000000)
	{
		nsec=(y->tv_usec-x->tv_usec)/1000000;
		y->tv_usec-=1000000*nsec;
		y->tv_sec+=nsec;
	}
	res->tv_sec=x->tv_sec-y->tv_sec;
	res->tv_usec=x->tv_usec-y->tv_usec;
}

/************************************************************************\
*
* Check for timeouts.
*
\************************************************************************/
void GtApp::CheckTimeOut(void)
{
	struct timeval ct;
	GtTOList *to,*tmp;
	GtDLink hd;
	
	if((to=(GtTOList *)GtDFirst(&tohd)))
	{
		GtDInit(&hd);
		gettimeofday(&ct,NULL);
		for(; to; to=tmp)
		{
			tmp=(GtTOList *)GtDNext(&tohd,&to->lnk);
			if(timercmp(&to->tv,&ct,<))
			{
				GtDRemove(&to->lnk);
				GtDAddTail(&hd,&to->lnk);
			}
		}
		while((to=(GtTOList *)GtDRemHead(&hd)))
		{
			to->Fun(this,NULL,to->ud);
			GtDAddTail(&ftohd,&to->lnk);
		}
	}
}
